import importlib
import json
import logging

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from django.conf import settings
from django.http.response import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from . exceptions import FatalException

logger = logging.getLogger(__name__)


@csrf_exempt
def ExampleAPI(request, schema, example):

    response = _example_api(request, schema, example)
    if isinstance(response, HttpResponse):
        return response
    else:
        return HttpResponse(response)


def _example_api(request, schema, example):

    response = None
    if schema:
        # If there is a problem with the json data, return a 400.
        try:
            data = json.loads(request.body.decode("utf-8"))
            validate(data, schema)
        except Exception as e:
            if hasattr(settings, 'RAMLWRAP_VALIDATION_ERROR_HANDLER') and settings.RAMLWRAP_VALIDATION_ERROR_HANDLER:
                response = _call_custom_handler(e)
            else:
                response = _validation_error_handler(e)

    if response:
        return response

    if not example:
        return None
    else:
        return example


def _is_valid_query(params, expected_params):
    """
    Function to validate get request params.
    """

    # If expected params, check them. If not, pass.
    if expected_params:
        for param in expected_params:
            # If the expected param is in the query.
            if param in params:
                for check, rule in expected_params[param].__dict__.items():
                    if rule is not None:
                        error_message = "QueryParam [%s] failed validation check [%s]:[%s]" % (param, check, rule)
                        if check == "minLength":
                            if len(params.get(param)) < rule:
                                raise ValidationError(error_message)
                        elif check == "maxLength":
                            if len(params.get(param)) > rule:
                                raise ValidationError(error_message)
            # Isn't in the query but it is required, throw a validation exception.
            elif expected_params[param].required is True:
                raise ValidationError("QueryParam [%s] failed validation check [Required]:[True]" % param)

    # TODO Add more checks here.
    return True


@csrf_exempt
def ValidatedGETAPI(request, expected_params, target):
    """
    Validate GET APIs.
    """

    if _is_valid_query(request.GET, expected_params):
        response = target(request)

        if isinstance(response, HttpResponse):
            return response
        else:
            return HttpResponse(json.dumps(response))


@csrf_exempt
def ValidatedPOSTAPI(request, schema, expected_params, target):
    """
    Validate POST APIs.
    """

    error_response = None
    if expected_params:
        _is_valid_query(request.GET, expected_params)   # Either passes through or raises an exception.

    if schema:
        # If there is a problem with the json data, return a 400.
        try:
            data = json.loads(request.body.decode("utf-8"))
            validate(data, schema) 
        except Exception as e:
            # Check the value is in settings, and that it is not None
            if hasattr(settings, 'RAMLWRAP_VALIDATION_ERROR_HANDLER') and settings.RAMLWRAP_VALIDATION_ERROR_HANDLER:
                error_response = _call_custom_handler(e)

            else:
                error_response = _validation_error_handler(e)
    else:
        data = json.loads(request.body.decode('utf-8'))

    if error_response:
        response = error_response
    else:
        request.validated_data = data
        response = target(request)

    if isinstance(response, HttpResponse):
        return response
    else:
        return HttpResponse(json.dumps(response))


def _validation_error_handler(e):
    """
    Handle validation errors.
    This can be overridden via the settings.
    """

    if isinstance(e, ValidationError):
        message = "Validation failed. {}".format(e.message)
        error_response = {
            "message": message,
            "code": e.validator
        }
        logger.info(message)
        error_resp = HttpResponse(error_response, status=422)
    else:
        raise FatalException("Malformed JSON in the request.", 400)

    return error_resp


def _call_custom_handler(e):
    """
    Dynamically import and call the custom handler
    defined by the user in the django settings file.
    """

    handler_full_path = settings.RAMLWRAP_VALIDATION_ERROR_HANDLER
    handler_method = handler_full_path.split(".")[-1]
    handler_class_path = ".".join(handler_full_path.split(".")[0:-1])

    handler = getattr(importlib.import_module(handler_class_path), handler_method)
    return handler(e)



"""
The Swagger parsing section of Raml Wrap
"""

import json
import logging

from swagger_parser import SwaggerParser
from django.conf.urls import url

from . validation import ExampleAPI, ValidatedPOSTAPI, ValidatedGETAPI

logger = logging.getLogger(__name__)


def swagger_url_patterns(swagger_dict, function_map):

    # Init with dictionary
    parser = SwaggerParser(swagger_dict=swagger_dict)

    tree = parser

    resource_map = {}

    _parse_swagger_child(tree, resource_map)

    patterns = _generate_swagger_patterns(resource_map, function_map, tree)

    return patterns


def _parse_swagger_child(resource, resource_map):

    # This has been carefully designed so as to not use recursion.
    if resource.paths is not None:
        for next in resource.paths:
            res = resource.paths[next]
            res["path"] = "%s%s" % ("", str(next))
            resource_map[res["path"]] = res


def _generate_swagger_patterns(resource_map, function_map, tree):

    patterns = []

    example_json = tree.definitions_example

    for t_url, resource in resource_map.items():

        t_url = t_url[1:]   # String leading /
        example = None

        if resource["path"] is not None:
            schema = None
            if "post" in resource:
                expected_params = None
                if "parameters" in resource["post"]:
                    expected_params = dict(resource["post"]["parameters"])
                    # Get the name from the parameters so we can grab the schema
                    for name in resource["post"]["parameters"]:
                        if "schema" in resource["post"]["parameters"][name]:
                            definition = resource["post"]["parameters"][name]["schema"]["$ref"]
                            # split off the definition name
                            definition = definition.split('#/definitions/')[1]
                            schema = json.loads(tree.json_specification)["definitions"][definition]

                        if "200" in resource["post"]["responses"]:
                            if "schema" in resource["post"]["responses"]["200"]:
                                definition = resource["post"]["responses"]["200"]["schema"]["$ref"]
                                # split off the definition name
                                definition = definition.split('#/definitions/')[1]
                                if definition in example_json:
                                    example = example_json[definition]

                if t_url in function_map:
                    patterns.append(url("^%s$" % t_url, ValidatedPOSTAPI, {'target': function_map[t_url], 'schema': schema, 'expected_params': expected_params}))
                else:
                    patterns.append(url("^%s$" % t_url, ExampleAPI, {'example': example, 'schema': schema}))

            if "get" in resource:
                expected_params = None
                if "parameters" in resource["get"]:
                    expected_params = dict(resource["get"]["parameters"])
                if "responses" in resource["get"]:
                    if "200" in resource["get"]["responses"]:
                        if "schema" in resource["get"]["responses"]["200"]:
                            definition = resource["get"]["responses"]["200"]["schema"]["$ref"]
                            # split off the definition name
                            definition = definition.split('#/definitions/')[1]
                            if definition in example_json:
                                example = example_json[definition]

                if t_url in function_map:
                    patterns.append(url("^%s$" % t_url, ValidatedGETAPI, {'target': function_map[t_url], 'expected_params': expected_params}))
                else:
                    patterns.append(url("^%s$" % t_url, ExampleAPI, {'example': example, 'schema': schema}))

    return patterns


